@@ -126,7 +126,7 @@ static int set_bad_null_error(Field *field, int err)
     return 0;
   case CHECK_FIELD_ERROR_FOR_NULL:
     if (!field->table->in_use->no_errors)
-      my_error(ER_BAD_NULL_ERROR, MYF(0), field->field_name.str);
+      my_error(err, MYF(0), field->field_name.str);
     return -1;
   }
   DBUG_ASSERT(0); // impossible
@@ -164,7 +164,7 @@ int set_field_to_null(Field *field)
     If no_conversion was not set, an error message is printed
 */
 
-int convert_null_to_field_value_or_error(Field *field)
+int convert_null_to_field_value_or_error(Field *field, uint err)
 {
   if (field->type() == MYSQL_TYPE_TIMESTAMP)
   {
@@ -172,14 +172,16 @@ int convert_null_to_field_value_or_error(Field *field)
     return 0;
   }
 
-  field->reset(); // Note: we ignore any potential failure of reset() here.
+  MY_BITMAP *old_map= dbug_tmp_use_all_columns(field->table, &field->table->write_set);
+  field->make_empty_rec_reset();
+  dbug_tmp_restore_column_map(&field->table->write_set, old_map);
 
   if (field == field->table->next_number_field)
   {
     field->table->auto_increment_field_not_null= FALSE;
     return 0;                             // field is set in fill_record()
   }
-  return set_bad_null_error(field, ER_BAD_NULL_ERROR);
+  return set_bad_null_error(field, err);
 }
 
 /**
@@ -216,7 +218,7 @@ set_field_to_null_with_conversions(Field *field, bool no_conversions)
   if (no_conversions)
     return -1;
 
-  return convert_null_to_field_value_or_error(field);
+  return convert_null_to_field_value_or_error(field, ER_BAD_NULL_ERROR);
 }
 
 
@@ -369,7 +371,7 @@ void Field::do_field_string(Copy_field *copy)
   res.length(0U);
 
   copy->from_field->val_str(&res);
-  copy->to_field->store(res.c_ptr_quick(), res.length(), res.charset());
+  copy->to_field->store(res.ptr(), res.length(), res.charset());
 }
 
 
@@ -385,14 +387,14 @@ void Field_enum::do_field_enum(Copy_field *copy)
 static void do_field_varbinary_pre50(Copy_field *copy)
 {
   char buff[MAX_FIELD_WIDTH];
-  copy->tmp.set_quick(buff,sizeof(buff),copy->tmp.charset());
+  copy->tmp.set_buffer_if_not_allocated(buff,sizeof(buff),copy->tmp.charset());
   copy->from_field->val_str(&copy->tmp);
 
   /* Use the same function as in 4.1 to trim trailing spaces */
-  size_t length= my_lengthsp_8bit(&my_charset_bin, copy->tmp.c_ptr_quick(),
+  size_t length= my_lengthsp_8bit(&my_charset_bin, copy->tmp.ptr(),
                                 copy->from_field->field_length);
 
-  copy->to_field->store(copy->tmp.c_ptr_quick(), length,
+  copy->to_field->store(copy->tmp.ptr(), length,
                         copy->tmp.charset());
 }
 
