@@ -68,7 +68,7 @@ static bool admin_recreate_table(THD *thd, TABLE_LIST *table_list,
 
   DEBUG_SYNC(thd, "ha_admin_try_alter");
   tmp_disable_binlog(thd); // binlogging is done by caller if wanted
-  result_code= (thd->open_temporary_tables(table_list) ||
+  result_code= (thd->check_and_open_tmp_table(table_list) ||
                 mysql_recreate_table(thd, table_list, recreate_info, false));
   reenable_binlog(thd);
   /*
@@ -565,6 +565,9 @@ static bool mysql_admin_table(THD* thd, TABLE_LIST* tables,
   Disable_wsrep_on_guard wsrep_on_guard(thd, disable_wsrep_on);
 #endif /* WITH_WSREP */
 
+  if (thd->transaction->xid_state.check_has_uncommitted_xa())
+    DBUG_RETURN(TRUE);
+
   fill_check_table_metadata_fields(thd, &field_list);
 
   if (protocol->send_result_set_metadata(&field_list,
@@ -987,8 +990,9 @@ static bool mysql_admin_table(THD* thd, TABLE_LIST* tables,
               types here.
             */
             enum enum_field_types type= field->type();
-            if (type < MYSQL_TYPE_TINY_BLOB ||
-                type > MYSQL_TYPE_BLOB)
+            if ((type < MYSQL_TYPE_TINY_BLOB ||
+                 type > MYSQL_TYPE_BLOB) &&
+                field->field_length <= thd->variables.analyze_max_length)
             {
               field->register_field_in_read_map();
               bitmap_set_bit(&tab->has_value_set, field->field_index);
@@ -1447,8 +1451,7 @@ send_result_message:
   }
   thd->resume_subsequent_commits(suspended_wfc);
   DBUG_EXECUTE_IF("inject_analyze_table_sleep", my_sleep(500000););
-  if (is_table_modified && is_cmd_replicated &&
-      (!opt_readonly || thd->slave_thread) && !thd->lex->no_write_to_binlog)
+  if (is_table_modified && is_cmd_replicated && !thd->lex->no_write_to_binlog)
   {
     thd->get_stmt_da()->set_overwrite_status(true);
     auto res= write_bin_log(thd, true, thd->query(), thd->query_length());
@@ -1585,9 +1588,14 @@ bool Sql_cmd_check_table::execute(THD *thd)
   bool res= TRUE;
   DBUG_ENTER("Sql_cmd_check_table::execute");
 
-  if (check_table_access(thd, SELECT_ACL, first_table,
-                         TRUE, UINT_MAX, FALSE))
+  if (check_some_access(thd, TABLE_ACLS, first_table))
+  {
+    my_error(ER_TABLEACCESS_DENIED_ERROR, MYF(0), "CHECK TABLE",
+            thd->security_ctx->priv_user,
+            thd->security_ctx->host_or_ip,
+            first_table->db.str, first_table->table_name.str);
     goto error; /* purecov: inspected */
+  }
 
   res= mysql_admin_table(thd, first_table, &m_lex->check_opt, &msg_check,
                          lock_type, 0, 0, HA_OPEN_FOR_REPAIR, 0,
