@@ -16,11 +16,11 @@
 #include "mariadb.h"
 #include "mysql/service_wsrep.h"
 #include "wsrep_binlog.h"
-#include "wsrep_priv.h"
 #include "log.h"
 #include "slave.h"
 #include "log_event.h"
 #include "wsrep_applier.h"
+#include "wsrep_mysqld.h"
 
 #include "transaction.h"
 
@@ -118,12 +118,13 @@ cleanup:
  */
 static int wsrep_write_cache_inc(THD*      const thd,
                                  IO_CACHE* const cache,
+                                 size_t    const log_position,
                                  size_t*   const len)
 {
   DBUG_ENTER("wsrep_write_cache_inc");
   my_off_t const saved_pos(my_b_tell(cache));
 
-  if (reinit_io_cache(cache, READ_CACHE, thd->wsrep_sr().log_position(), 0, 0))
+  if (reinit_io_cache(cache, READ_CACHE, log_position, 0, 0))
   {
     WSREP_ERROR("failed to initialize io-cache");
     DBUG_RETURN(1);;
@@ -157,7 +158,7 @@ static int wsrep_write_cache_inc(THD*      const thd,
     } while ((cache->file >= 0) && (length= my_b_fill(cache)));
     if (ret == 0)
     {
-      assert(total_length + thd->wsrep_sr().log_position() == saved_pos);
+      assert(total_length + log_position == saved_pos);
     }
   }
 
@@ -178,9 +179,10 @@ cleanup:
  */
 int wsrep_write_cache(THD*      const thd,
                       IO_CACHE* const cache,
+                      size_t    const log_position,
                       size_t*   const len)
 {
-  return wsrep_write_cache_inc(thd, cache, len);
+  return wsrep_write_cache_inc(thd, cache, log_position, len);
 }
 
 void wsrep_dump_rbr_buf(THD *thd, const void* rbr_buf, size_t buf_len)
@@ -198,7 +200,7 @@ void wsrep_dump_rbr_buf(THD *thd, const void* rbr_buf, size_t buf_len)
     to alloc and pass as an argument to snprintf.
   */
 
-  char *filename= (char *)malloc(len+1);
+  char *filename= (char *) my_malloc(key_memory_WSREP, len+1, 0);
   int len1= snprintf(filename, len+1, "%s/GRA_%lld_%lld.log",
                     wsrep_data_home_dir, (longlong) thd->thread_id,
                     (long long)wsrep_thd_trx_seqno(thd));
@@ -206,7 +208,7 @@ void wsrep_dump_rbr_buf(THD *thd, const void* rbr_buf, size_t buf_len)
   if (len > len1)
   {
     WSREP_ERROR("RBR dump path truncated: %d, skipping dump.", len);
-    free(filename);
+    my_free(filename);
     return;
   }
 
@@ -225,7 +227,7 @@ void wsrep_dump_rbr_buf(THD *thd, const void* rbr_buf, size_t buf_len)
     WSREP_ERROR("Failed to open file '%s': %d (%s)",
                 filename, errno, strerror(errno));
   }
-  free(filename);
+  my_free(filename);
 }
 
 /* Dump replication buffer along with header to a file. */
@@ -248,7 +250,7 @@ void wsrep_dump_rbr_buf_with_header(THD *thd, const void *rbr_buf,
     to alloc and pass as an argument to snprintf.
   */
   char *filename;
-  if (len < 0 || !(filename= (char*)malloc(len+1)))
+  if (len < 0 || !(filename= (char*) my_malloc(key_memory_WSREP, len+1, 0)))
   {
     WSREP_ERROR("snprintf error: %d, skipping dump.", len);
     DBUG_VOID_RETURN;
@@ -261,7 +263,7 @@ void wsrep_dump_rbr_buf_with_header(THD *thd, const void *rbr_buf,
   if (len > len1)
   {
     WSREP_ERROR("RBR dump path truncated: %d, skipping dump.", len);
-    free(filename);
+    my_free(filename);
     DBUG_VOID_RETURN;
   }
 
@@ -301,7 +303,7 @@ cleanup2:
   end_io_cache(&cache);
 
 cleanup1:
-  free(filename);
+  my_free(filename);
   mysql_file_close(file, MYF(MY_WME));
 
   if (!thd->wsrep_applier) delete ev;
