@@ -2022,7 +2022,7 @@ static void fil_crypt_thread()
 #endif /* UNIV_PFS_THREAD */
 	mysql_mutex_lock(&fil_crypt_threads_mutex);
 	rotate_thread_t thr(srv_n_fil_crypt_threads_started++);
-	pthread_cond_signal(&fil_crypt_cond); /* signal that we started */
+	pthread_cond_broadcast(&fil_crypt_cond);
 
 	if (!thr.should_shutdown()) {
 		/* if we find a tablespace that is starting, skip over it
@@ -2094,7 +2094,7 @@ wait_for_work:
 
 	fil_crypt_return_iops(&thr);
 	srv_n_fil_crypt_threads_started--;
-	pthread_cond_signal(&fil_crypt_cond); /* signal that we stopped */
+	pthread_cond_broadcast(&fil_crypt_cond);
 	mysql_mutex_unlock(&fil_crypt_threads_mutex);
 
 	my_thread_end();
@@ -2108,6 +2108,9 @@ Adjust thread count for key rotation
 @param[in]	enw_cnt		Number of threads to be used */
 void fil_crypt_set_thread_cnt(const uint new_cnt)
 {
+	if (srv_read_only_mode)
+		return;
+
 	if (!fil_crypt_threads_inited) {
 		if (srv_shutdown_state != SRV_SHUTDOWN_NONE)
 			return;
@@ -2261,6 +2264,8 @@ void fil_crypt_set_encrypt_tables(ulong val)
 Init threads for key rotation */
 void fil_crypt_threads_init()
 {
+	ut_ad(!srv_read_only_mode);
+
 	if (!fil_crypt_threads_inited) {
 		pthread_cond_init(&fil_crypt_cond, nullptr);
 		pthread_cond_init(&fil_crypt_threads_cond, nullptr);
