@@ -354,12 +354,16 @@ static bool fil_node_open_file_low(fil_node_t *node, const byte *page,
     bool success;
     node->handle= os_file_create(innodb_data_file_key, node->name,
                                  node->is_raw_disk
-                                 ? OS_FILE_OPEN_RAW : OS_FILE_OPEN,
+                                 ? OS_FILE_OPEN_RAW
+                                 : node->deferred
+                                 ? OS_FILE_OPEN_SILENT : OS_FILE_OPEN,
                                  type,
                                  srv_read_only_mode, &success);
+    ut_ad(success == node->is_open());
 
     if (success && node->is_open())
     {
+    created:
 #ifndef _WIN32
       if (!node->space->id && !srv_read_only_mode && my_disable_locking &&
           os_file_lock(node->handle, node->name))
@@ -371,6 +375,15 @@ static bool fil_node_open_file_low(fil_node_t *node, const byte *page,
 #endif
       break;
     }
+    else if (node->deferred)
+    {
+      node->handle= os_file_create(innodb_data_file_key, node->name,
+                                   OS_FILE_CREATE, type,
+                                   srv_read_only_mode, &success);
+      ut_ad(success == node->is_open());
+      if (node->is_open())
+        goto created;
+    }
 
     /* The following call prints an error message */
     if (os_file_get_last_error(true) == EMFILE + 100 &&
@@ -860,13 +873,7 @@ static void fil_space_free_low(fil_space_t *space) noexcept
 	/* The tablespace must not be in fil_system.named_spaces. */
 	ut_ad(srv_fast_shutdown == 2 || !srv_was_started
 	      || space->max_lsn == 0);
-
-	/* Wait for fil_space_t::release() after
-	fil_system_t::detach(), the tablespace cannot be found, so
-	fil_space_t::get() would return NULL */
-	while (space->referenced()) {
-		std::this_thread::sleep_for(std::chrono::microseconds(100));
-	}
+	ut_ad(!space->referenced());
 
 	for (fil_node_t* node = UT_LIST_GET_FIRST(space->chain);
 	     node != NULL; ) {
@@ -921,7 +928,7 @@ bool fil_space_free(uint32_t id, bool x_latched) noexcept
 		} else {
 			ut_ad(log_sys.latch_have_wr());
 			if (space->max_lsn) {
-				ut_d(space->max_lsn = 0);
+				space->max_lsn = 0;
 				fil_system.named_spaces.remove(*space);
 			}
 		}
@@ -952,8 +959,6 @@ fil_space_t *fil_space_t::create(uint32_t id, uint32_t flags,
   ut_ad(fil_space_t::is_valid_flags(flags & ~FSP_FLAGS_MEM_MASK, id));
   ut_ad(srv_page_size == UNIV_PAGE_SIZE_ORIG || flags != 0);
 
-  DBUG_EXECUTE_IF("fil_space_create_failure", return nullptr;);
-
   fil_space_t** after= fil_system.spaces.cell_get(id)->search
     (&fil_space_t::hash, [id](const fil_space_t *space)
     { return !space || space->id == id; });
@@ -1423,12 +1428,15 @@ fil_space_t *fil_space_t::get(uint32_t id) noexcept
 @param type           file operation
 @param first_page_no  first page number in the file
 @param path           file path
-@param new_path       new file path for type=FILE_RENAME */
-inline void mtr_t::log_file_op(mfile_type_t type, uint32_t space_id,
-			       const char *path, const char *new_path)
+@param new_path       new file path for type=FILE_RENAME
+@return number of bytes written */
+inline size_t mtr_t::log_file_op(mfile_type_t type, uint32_t space_id,
+                                 const char *path, const char *new_path)
+  noexcept
 {
   ut_ad((new_path != nullptr) == (type == FILE_RENAME));
   ut_ad(!(byte(type) & 15));
+  ut_ad(!is_predefined_tablespace(space_id));
 
   /* fil_name_parse() requires that there be at least one path
   separator and that the file path end with ".ibd". */
@@ -1437,7 +1445,7 @@ inline void mtr_t::log_file_op(mfile_type_t type, uint32_t space_id,
 
   m_modifications= true;
   if (!is_logged())
-    return;
+    return 0;
   m_last= nullptr;
 
   const size_t len= strlen(path);
@@ -1469,7 +1477,7 @@ inline void mtr_t::log_file_op(mfile_type_t type, uint32_t space_id,
 
   m_log.close(end);
 
-  if (new_path)
+  if (new_len)
   {
     ut_ad(strchr(new_path, '/'));
     m_log.push(reinterpret_cast<const byte*>(path), uint32_t(len + 1));
@@ -1477,17 +1485,8 @@ inline void mtr_t::log_file_op(mfile_type_t type, uint32_t space_id,
   }
   else
     m_log.push(reinterpret_cast<const byte*>(path), uint32_t(len));
-}
 
-/** Write FILE_MODIFY for a file.
-@param[in]	space_id	tablespace id
-@param[in]	name		tablespace file name
-@param[in,out]	mtr		mini-transaction */
-static void fil_name_write(uint32_t space_id, const char *name,
-                           mtr_t *mtr)
-{
-  ut_ad(!is_predefined_tablespace(space_id));
-  mtr->log_file_op(FILE_MODIFY, space_id, name);
+  return end - log_ptr + len + new_len;
 }
 
 fil_space_t *fil_space_t::drop(uint32_t id, pfs_os_file_t *detached_handle)
@@ -1593,10 +1592,24 @@ fil_space_t *fil_space_t::drop(uint32_t id, pfs_os_file_t *detached_handle)
 
   pfs_os_file_t handle= fil_system.detach(space, true);
   mysql_mutex_unlock(&fil_system.mutex);
+  /* The above mtr.commit_file(*space, nullptr) should remove the space from
+  fil_system.named_spaces. Before we set the STOPPING_WRITES flag, another
+  concurrent operation could have marked the tablespace dirty again.
+  This clean-up corresponds to fil_space_free(). */
+  log_sys.latch.wr_lock(SRW_LOCK_CALL);
+  ut_ad((space->pending() & ~NEEDS_FSYNC) == (STOPPING | CLOSING));
+  if (space->max_lsn != 0)
+  {
+    space->max_lsn= 0;
+    fil_system.named_spaces.remove(*space);
+  }
+  log_sys.latch.wr_unlock();
+
   if (detached_handle)
     *detached_handle = handle;
   else
     os_file_close(handle);
+
   return space;
 }
 
@@ -1622,12 +1635,6 @@ void fil_close_tablespace(uint32_t id) noexcept
 	while (buf_flush_list_space(space));
 
 	space->x_unlock();
-	log_sys.latch.wr_lock(SRW_LOCK_CALL);
-	if (space->max_lsn != 0) {
-		ut_d(space->max_lsn = 0);
-		fil_system.named_spaces.remove(*space);
-	}
-	log_sys.latch.wr_unlock();
 	fil_space_free_low(space);
 }
 
@@ -1849,16 +1856,12 @@ fil_ibd_create(
 	uint32_t	key_id,
 	dberr_t*	err) noexcept
 {
-	pfs_os_file_t	file;
-	bool		success;
-	mtr_t		mtr;
-	bool		has_data_dir = FSP_FLAGS_HAS_DATA_DIR(flags) != 0;
-
 	ut_ad(!is_system_tablespace(space_id));
 	ut_ad(!srv_read_only_mode);
 	ut_a(space_id < SRV_SPACE_ID_UPPER_BOUND);
 	ut_a(size >= FIL_IBD_FILE_INITIAL_SIZE);
-	ut_a(fil_space_t::is_valid_flags(flags & ~FSP_FLAGS_MEM_MASK, space_id));
+	ut_a(fil_space_t::is_valid_flags(flags & ~FSP_FLAGS_MEM_MASK,
+					 space_id));
 
 	/* Create the subdirectories in the path, if they are
 	not there already. */
@@ -1867,14 +1870,6 @@ fil_ibd_create(
 		return NULL;
 	}
 
-	mtr.start();
-	mtr.log_file_op(FILE_CREATE, space_id, path);
-	log_sys.latch.wr_lock(SRW_LOCK_CALL);
-	auto lsn= mtr.commit_files();
-	log_sys.latch.wr_unlock();
-	mtr.flag_wr_unlock();
-	log_write_up_to(lsn, true);
-
 	static_assert(((UNIV_ZIP_SIZE_MIN >> 1) << 3) == 4096,
 		      "compatibility");
 #if defined _WIN32 || defined O_DIRECT
@@ -1890,45 +1885,7 @@ fil_ibd_create(
 #else
 	constexpr auto type = OS_DATA_FILE;
 #endif
-
-	file = os_file_create(
-		innodb_data_file_key, path,
-		OS_FILE_CREATE,
-		type, srv_read_only_mode, &success);
-
-	if (!success) {
-		/* The following call will print an error message */
-		switch (os_file_get_last_error(true)) {
-		case OS_FILE_ALREADY_EXISTS:
-			ib::info() << "The file '" << path << "'"
-				" already exists though the"
-				" corresponding table did not exist"
-				" in the InnoDB data dictionary."
-				" You can resolve the problem by removing"
-				" the file.";
-			*err = DB_TABLESPACE_EXISTS;
-			break;
-		case OS_FILE_DISK_FULL:
-			*err = DB_OUT_OF_FILE_SPACE;
-			break;
-		default:
-			*err = DB_ERROR;
-		}
-		ib::error() << "Cannot create file '" << path << "'";
-		return NULL;
-	}
-
 	const bool is_compressed = fil_space_t::is_compressed(flags);
-#ifdef _WIN32
-	const bool is_sparse = is_compressed;
-	if (is_compressed) {
-		os_file_set_sparse_win32(file);
-	}
-#else
-	const bool is_sparse = is_compressed
-		&& DB_SUCCESS == os_file_punch_hole(file, 0, 4096)
-		&& !my_test_if_thinly_provisioned(file);
-#endif
 
 	if (fil_space_t::full_crc32(flags)) {
 		flags |= FSP_FLAGS_FCRC32_PAGE_SSIZE();
@@ -1943,54 +1900,122 @@ fil_ibd_create(
 		? fil_space_create_crypt_data(mode, key_id)
 		: nullptr;
 
+	mysql_mutex_lock(&fil_system.mutex);
+	fil_space_t* space = fil_space_t::create(space_id, flags, false,
+						 crypt_data, mode, true);
+	fil_node_t* node = space->add(path, OS_FILE_CLOSED, size, false, true);
+	space->set_stopped();
+	mysql_mutex_unlock(&fil_system.mutex);
+
+	buf_block_t *header[2];
+	{
+		mtr_t mtr;
+		mtr.start();
+		mtr.log_file_op(FILE_CREATE, space_id, path);
+		mtr.set_named_space(space);
+		log_free_check();
+		ut_a(fsp_header_init(space, size, &mtr) == DB_SUCCESS);
+		ut_ad(mtr.get_savepoint() == 3);
+		header[0] = mtr.at_savepoint(1);
+		header[1] = mtr.at_savepoint(2);
+		ut_ad(header[0]->page.id() == page_id_t(space_id, 0));
+		ut_ad(header[1]->page.id() == page_id_t(space_id, 1));
+		/* Block any page writes before the file has been created. */
+		header[0]->page.lock.x_lock();
+		header[1]->page.lock.x_lock();
+		mtr.commit();
+		/* Durably write the FILE_CREATE record (as well as records for
+		the fsp_header_init() above before creating the file. */
+		log_write_up_to(mtr.commit_lsn(), true);
+	}
+
+	DEBUG_SYNC_C("fil_ibd_create_logged");
+
+	bool success;
+	pfs_os_file_t file = os_file_create(
+		innodb_data_file_key, path, OS_FILE_CREATE,
+		type, srv_read_only_mode, &success);
+	ut_ad(!success == (file == OS_FILE_CLOSED));
+
+	if (!success) {
+		*err = DB_ERROR;
+		/* The following call will print an error message */
+		switch (os_file_get_last_error(true)) {
+		case OS_FILE_ALREADY_EXISTS:
+			sql_print_error("InnoDB: The file '%s'"
+					" already exists though the"
+					" corresponding table did not exist"
+					" in the InnoDB data dictionary."
+					" You can resolve the problem"
+					" by removing"
+					" the file.", path);
+			*err = DB_TABLESPACE_EXISTS;
+			break;
+		case OS_FILE_DISK_FULL:
+			*err = DB_OUT_OF_FILE_SPACE;
+			/* fall through */
+		default:
+			sql_print_error("InnoDB: Cannot create file '%s'",
+					path);
+		}
+		goto err_exit_after_cleanup;
+err_exit:
+		os_file_delete(innodb_data_file_key, path);
+		os_file_close(file);
+err_exit_after_cleanup:
+		space->release();
+		fil_space_free(space_id, false);
+		space = nullptr;
+func_exit:
+		header[0]->page.lock.x_unlock();
+		header[1]->page.lock.x_unlock();
+		DBUG_EXECUTE_IF("checkpoint_after_file_create",
+				log_make_checkpoint(););
+		return space;
+	}
+
+#ifdef _WIN32
+	const bool is_sparse = is_compressed;
+	if (is_compressed) {
+		os_file_set_sparse_win32(file);
+	}
+#else
+	const bool is_sparse = is_compressed
+		&& DB_SUCCESS == os_file_punch_hole(file, 0, 4096)
+		&& !my_test_if_thinly_provisioned(file);
+#endif
 	if (!os_file_set_size(path, file,
 			      os_offset_t(size) << srv_page_size_shift,
 			      is_sparse)) {
 		*err = DB_OUT_OF_FILE_SPACE;
-err_exit:
-		os_file_close(file);
-		os_file_delete(innodb_data_file_key, path);
-		free(crypt_data);
-		return nullptr;
+		goto err_exit;
 	}
 
-	fil_space_t::name_type space_name;
+	DBUG_EXECUTE_IF("fil_space_create_failure",
+			*err = DB_OUT_OF_FILE_SPACE; goto err_exit;);
 
-	if (has_data_dir) {
+	if (FSP_FLAGS_HAS_DATA_DIR(flags)) {
 		/* Make the ISL file if the IBD file is not
 		in the default location. */
-		space_name = {name.m_name, strlen(name.m_name)};
+		fil_space_t::name_type space_name{name.m_name,
+						  strlen(name.m_name)};
 		*err = RemoteDatafile::create_link_file(space_name, path);
 		if (*err != DB_SUCCESS) {
 			goto err_exit;
 		}
 	}
 
-	DBUG_EXECUTE_IF("checkpoint_after_file_create",
-			log_make_checkpoint(););
-
 	mysql_mutex_lock(&fil_system.mutex);
-	if (fil_space_t* space = fil_space_t::create(space_id,
-						     flags, false,
-						     crypt_data, mode, true)) {
-		fil_node_t* node = space->add(path, file, size, false, true);
-		node->find_metadata(IF_WIN(,true));
-		mysql_mutex_unlock(&fil_system.mutex);
-		mtr.start();
-		mtr.set_named_space(space);
-		ut_a(fsp_header_init(space, size, &mtr) == DB_SUCCESS);
-		mtr.commit();
-		return space;
-	} else {
-		mysql_mutex_unlock(&fil_system.mutex);
+	space->clear_stopped();
+	node->handle = file;
+	node->find_metadata(IF_WIN(,true));
+	if (++fil_system.n_open >= srv_max_n_open_files) {
+		space->reacquire();
+		fil_space_t::try_to_close(space, true);
+		space->release();
 	}
-
-	if (space_name.data()) {
-		RemoteDatafile::delete_link_file(space_name);
-	}
-
-	*err = DB_ERROR;
-	goto err_exit;
+	mysql_mutex_unlock(&fil_system.mutex);
+	goto func_exit;
 }
 
 fil_space_t *fil_ibd_open(uint32_t id, uint32_t flags,
@@ -2174,7 +2199,9 @@ func_exit:
 			    || df_remote.is_open() != df_remote.is_valid()) {
 				goto corrupted;
 			}
+#ifndef DBUG_OFF
 error:
+#endif
 			local_err = DB_ERROR;
 			goto func_exit;
 		}
@@ -2198,6 +2225,8 @@ error:
 	ut_a(valid_tablespaces_found == 1);
 
 skip_validate:
+	DBUG_EXECUTE_IF("fil_space_create_failure", goto error;);
+
 	const byte* first_page =
 		df_default.is_open() ? df_default.get_first_page() :
 		df_remote.get_first_page();
@@ -2211,11 +2240,6 @@ skip_validate:
 	space = fil_space_t::create(id, flags,
 				    validate == fil_space_t::VALIDATE_IMPORT,
 				    crypt_data);
-	if (!space) {
-		mysql_mutex_unlock(&fil_system.mutex);
-		goto error;
-	}
-
 	/* We do not measure the size of the file, that is why
 	we pass the 0 below */
 
@@ -2517,11 +2541,6 @@ tablespace_check:
 	space = fil_space_t::create(uint32_t(space_id), flags, false,
 				    crypt_data);
 
-	if (space == NULL) {
-		mysql_mutex_unlock(&fil_system.mutex);
-		return(FIL_LOAD_INVALID);
-	}
-
 	ut_ad(space->id == file.space_id());
 	ut_ad(space->id == space_id);
 
@@ -2986,9 +3005,8 @@ ATTRIBUTE_NOINLINE ATTRIBUTE_COLD void mtr_t::name_write() noexcept
 
   mtr_t mtr;
   mtr.start();
-  fil_name_write(m_user_space->id,
-                 UT_LIST_GET_FIRST(m_user_space->chain)->name,
-                 &mtr);
+  mtr.log_file_op(FILE_MODIFY, m_user_space->id,
+                  UT_LIST_GET_FIRST(m_user_space->chain)->name);
   mtr.commit_files();
 }
 
@@ -3005,14 +3023,18 @@ ATTRIBUTE_COLD lsn_t fil_names_clear(lsn_t lsn) noexcept
 	ut_ad(log_sys.is_latest());
 
 	mtr.start();
+	constexpr size_t budget = recv_sys.MTR_SIZE_MAX - (3 + 5);
+	size_t budget_left = budget;
 
 	for (auto it = fil_system.named_spaces.begin();
 	     it != fil_system.named_spaces.end(); ) {
-		if (mtr.get_log_size() + strlen(it->chain.start->name)
-		    >= recv_sys.MTR_SIZE_MAX - (3 + 5)) {
+		const char* const name = it->chain.start->name;
+
+		if (strlen(name) >= budget_left) {
 			/* Prevent log parse buffer overflow */
 			mtr.commit_files();
 			mtr.start();
+			budget_left = budget;
 		}
 
 		auto next = std::next(it);
@@ -3033,8 +3055,9 @@ ATTRIBUTE_COLD lsn_t fil_names_clear(lsn_t lsn) noexcept
 		where max_lsn turned nonzero), we could avoid the
 		fil_names_write() call if min_lsn > lsn. */
 		ut_ad(UT_LIST_GET_LEN((*it).chain) == 1);
-		fil_name_write((*it).id, UT_LIST_GET_FIRST((*it).chain)->name,
-			       &mtr);
+		size_t s = mtr.log_file_op(FILE_MODIFY, (*it).id, name);
+		ut_ad(s <= budget_left);
+		budget_left -= s;
 		it = next;
 	}
 
