@@ -62,21 +62,21 @@ extern ulong		innodb_change_buffering;
 
 /** Insert buffer struct */
 struct ibuf_t{
-	Atomic_relaxed<ulint> size;	/*!< current size of the ibuf index
+	Atomic_relaxed<uint32_t> size;	/*!< current size of the ibuf index
 					tree, in pages */
-	Atomic_relaxed<ulint> max_size;	/*!< recommended maximum size of the
+	Atomic_relaxed<uint32_t> max_size;/*!< recommended maximum size of the
 					ibuf index tree, in pages */
-	ulint		seg_size;	/*!< allocated pages of the file
+	uint32_t	seg_size;	/*!< allocated pages of the file
 					segment containing ibuf header and
-					tree */
-	bool		empty;		/*!< Protected by the page
-					latch of the root page of the
-					insert buffer tree
-					(FSP_IBUF_TREE_ROOT_PAGE_NO). true
-					if and only if the insert
-					buffer tree is empty. */
-	ulint		free_list_len;	/*!< length of the free list */
-	ulint		height;		/*!< tree height */
+					tree; protected by ibuf.index->lock
+					and the root page latch */
+	bool		empty;		/*!< whether the change buffer is
+					empty; protected by ibuf.index->lock
+					and the root page latch */
+	uint8_t		height;		/*!< tree height */
+	uint32_t	free_list_len;	/*!< length of the free list;
+					protected by ibuf.index->lock and
+					the root page latch */
 	dict_index_t*	index;		/*!< insert buffer index */
 
 	/** number of pages merged */
@@ -279,8 +279,6 @@ Must not be called when recv_no_ibuf_operations==true.
 @param[in]	zip_size	ROW_FORMAT=COMPRESSED page size, or 0
 @param[in]	x_latch		FALSE if relaxed check (avoid latching the
 bitmap page)
-@param[in]	file		file name
-@param[in]	line		line where called
 @param[in,out]	mtr		mtr which will contain an x-latch to the
 bitmap page if the page is not one of the fixed address ibuf pages, or NULL,
 in which case a new transaction is created.
@@ -292,8 +290,6 @@ ibuf_page_low(
 #ifdef UNIV_DEBUG
 	bool			x_latch,
 #endif /* UNIV_DEBUG */
-	const char*		file,
-	unsigned		line,
 	mtr_t*			mtr)
 	MY_ATTRIBUTE((warn_unused_result));
 
@@ -305,7 +301,7 @@ Must not be called when recv_no_ibuf_operations==true.
 @param[in,out]	mtr		mini-transaction or NULL
 @return TRUE if level 2 or level 3 page */
 # define ibuf_page(page_id, zip_size, mtr)	\
-	ibuf_page_low(page_id, zip_size, true, __FILE__, __LINE__, mtr)
+	ibuf_page_low(page_id, zip_size, true, mtr)
 
 #else /* UNIV_DEBUG */
 
@@ -316,7 +312,7 @@ Must not be called when recv_no_ibuf_operations==true.
 @param[in,out]	mtr		mini-transaction or NULL
 @return TRUE if level 2 or level 3 page */
 # define ibuf_page(page_id, zip_size, mtr)	\
-	ibuf_page_low(page_id, zip_size, __FILE__, __LINE__, mtr)
+	ibuf_page_low(page_id, zip_size, mtr)
 
 #endif /* UNIV_DEBUG */
 /***********************************************************************//**
@@ -360,14 +356,16 @@ exist entries for such a page if the page belonged to an index which
 subsequently was dropped.
 @param block    X-latched page to try to apply changes to, or NULL to discard
 @param page_id  page identifier
-@param zip_size ROW_FORMAT=COMPRESSED page size, or 0 */
-void ibuf_merge_or_delete_for_page(buf_block_t *block, const page_id_t page_id,
-                                   ulint zip_size);
+@param zip_size ROW_FORMAT=COMPRESSED page size, or 0
+@return error code */
+dberr_t ibuf_merge_or_delete_for_page(buf_block_t *block,
+                                      const page_id_t page_id,
+                                      ulint zip_size);
 
 /** Delete all change buffer entries for a tablespace,
 in DISCARD TABLESPACE, IMPORT TABLESPACE, or read-ahead.
 @param[in]	space		missing or to-be-discarded tablespace */
-void ibuf_delete_for_discarded_space(ulint space);
+void ibuf_delete_for_discarded_space(uint32_t space);
 
 /** Contract the change buffer by reading pages to the buffer pool.
 @return a lower limit for the combined size in bytes of entries which
