@@ -39,10 +39,13 @@ static int s3_read_file_from_disk(const char *filename, uchar **to,
 
 /* Used by ha_s3.cc and tools to define different protocol options */
 
-static const char *protocol_types[]= {"Auto", "Original", "Amazon", NullS};
+static const char *protocol_types[]= {"Auto", "Original", "Amazon", "Legacy", "Path", "Domain", NullS};
 TYPELIB s3_protocol_typelib= {array_elements(protocol_types)-1,"",
                               protocol_types, NULL};
 
+static const char *providers[]= {"Default", "Amazon", "Huawei", NullS};
+TYPELIB s3_provider_typelib = {array_elements(providers)-1,"",providers, NULL};
+
 /******************************************************************************
  Allocations handler for libmarias3
  To be removed when we do the init allocation in mysqld.cc
@@ -154,15 +157,49 @@ ms3_st *s3_open_connection(S3_INFO *s3)
                     errno, ms3_error(errno));
     my_errno= HA_ERR_NO_SUCH_TABLE;
   }
-  if (s3->protocol_version)
+
+  /* Provider specific overrides */
+  switch (s3->provider)
+  {
+    case 0: /* Default */
+      break;
+    case 1: /* Amazon */
+      s3->protocol_version = 5;
+      break;
+    case 2: /* Huawei */
+      s3->no_content_type = 1;
+      break;
+  }
+
+  if (s3->protocol_version > 2)
+  {
+    uint8_t protocol_version;
+    switch (s3->protocol_version)
+    {
+      case 3: /* Legacy means v1 */
+      case 4: /* Path means v1 */
+        protocol_version= 1;
+        break;
+      case 5: /* Domain means v2 */
+        protocol_version= 2;
+        break;
+    }
+
     ms3_set_option(s3_client, MS3_OPT_FORCE_PROTOCOL_VERSION,
-                   &s3->protocol_version);
+                   &protocol_version);
+  }
   if (s3->port)
     ms3_set_option(s3_client, MS3_OPT_PORT_NUMBER, &s3->port);
 
   if (s3->use_http)
     ms3_set_option(s3_client, MS3_OPT_USE_HTTP, NULL);
 
+  if (s3->ssl_no_verify)
+    ms3_set_option(s3_client, MS3_OPT_DISABLE_SSL_VERIFY, NULL);
+
+  if (s3->no_content_type)
+    ms3_set_option(s3_client, MS3_OPT_NO_CONTENT_TYPE, NULL);
+
   return s3_client;
 }
 
@@ -351,7 +388,7 @@ int aria_copy_to_s3(ms3_st *s3_client, const char *aws_bucket,
       if (display)
         printf("Copying frm file %s\n", filename);
 
-      end= strmov(aws_path_end,"/frm");
+      strmov(aws_path_end,"/frm");
       convert_frm_to_s3_format(alloc_block);
 
       /* Note that frm is not compressed! */
@@ -374,7 +411,7 @@ int aria_copy_to_s3(ms3_st *s3_client, const char *aws_bucket,
                      O_RDONLY | O_SHARE | O_NOFOLLOW | O_CLOEXEC,
                      MYF(MY_WME))) < 0)
     DBUG_RETURN(1);
-  if ((error= aria_get_capabilities(file, &cap)))
+  if ((error= aria_get_capabilities(file, table_name, &cap)))
   {
     fprintf(stderr, "Got error %d when reading Aria header from %s\n",
             error, path);
@@ -1162,7 +1199,7 @@ my_bool s3_rename_object(ms3_st *s3_client, const char *aws_bucket,
       if (!(errmsg= ms3_server_error(s3_client)))
         errmsg= ms3_error(error);
 
-      my_printf_error(EE_READ, "Got error from move_object(%s -> %s): %d %",
+      my_printf_error(EE_READ, "Got error from move_object(%s -> %s): %d %s",
                       error_flags,
                       from_name, to_name, error, errmsg);
     }
@@ -1232,7 +1269,7 @@ static void convert_index_to_s3_format(uchar *header, ulong block_size,
   uchar *base_pos;
   uint  base_offset;
 
-  memcpy(state.header.file_version, header, sizeof(state.header));
+  memcpy(&state.header, header, sizeof(state.header));
   base_offset= mi_uint2korr(state.header.base_pos);
   base_pos= header + base_offset;
 
@@ -1251,7 +1288,7 @@ static void convert_index_to_disk_format(uchar *header)
   uchar *base_pos;
   uint  base_offset;
 
-  memcpy(state.header.file_version, header, sizeof(state.header));
+  memcpy(&state.header, header, sizeof(state.header));
   base_offset= mi_uint2korr(state.header.base_pos);
   base_pos= header + base_offset;
 
@@ -1310,7 +1347,7 @@ my_bool set_database_and_table_from_path(S3_INFO *s3, const char *path)
     if (path[length-1] == FN_LIBCHAR || path[length-1] == '/')
       break;
 #ifdef FN_DEVCHAR
-    if (path[length-1] == FN_DECVHAR)
+    if (path[length-1] == FN_DEVCHAR)
       break;
 #endif
   }
